/*------------------------------------------------------------------------*/
/* Copyright (C) 2018, Armin Biere, Johannes Kepler University Linz       */
/* Copyright (C) 2018, Daniela Ritirc, Johannes Kepler University Linz    */
/*------------------------------------------------------------------------*/



#define VERSION "000"

static const char * usage =
"drup2dpac [ <option> ... ]  [ <trace> <pac>]\n"
"\n"
"where <option> is one of the following\n"
"\n"
"  -h             print this command line option summary and exit\n"
"\n"
" <trace>      should point to a DRUP trace in TRACECHECK format\n"
" <pac>        should point to a DRUP trace in TRACECHECK format\n"
;

/*------------------------------------------------------------------------*/
// TODO delete unused rules, make counter for rules

#include <assert.h>
#include <ctype.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <gmp.h>

/*------------------------------------------------------------------------*/
static const char * input_file_name = 0;
static const char * input_file_name2 = 0;
static const char * output_file_name = 0;
static const char * output_file_name2 = 0;
static const char * output_file_name3 = 0;
static const char * output_file_name4 = 0;
static FILE * polys_file; //  .polys
static FILE * pac_file; //  .pac
static FILE * equal_file; //  .pac
static FILE * mapping_file;

static uint64_t start_traces = 0;
static uint64_t end_traces = 0;
static uint64_t bad_clause = 0;

static int64_t init_clauses = 0;
/*------------------------------------------------------------------------*/
static void die (const char * msg, ...) {
  fflush (stdout);
  fprintf (stderr, "*** 'drup2dpac' error: ");
  va_list ap;
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fprintf (stderr, "\n");
  fflush (stderr);
  exit (1);
}

static void msg (const char * msg, ...) {
  fprintf (stdout, "[drup2dpac] ");
  va_list ap;
  va_start (ap, msg);
  vfprintf (stdout, msg, ap);
  va_end (ap);
  fprintf (stdout, "\n");
  fflush (stdout);
}

/*------------------------------------------------------------------------*/
// memory management routines maintaining a count of allocated memory

static size_t current_allocated_bytes;
static size_t max_allocated_bytes;

static void inc_allocated_bytes (size_t bytes) {
  current_allocated_bytes += bytes;
  if (current_allocated_bytes > max_allocated_bytes)
    max_allocated_bytes = current_allocated_bytes;
}

static void dec_allocated_bytes (size_t bytes) {
  assert (current_allocated_bytes >= bytes);
  current_allocated_bytes -= bytes;
}

static void * allocate (size_t bytes) {
  if (!bytes) return 0;
  void * res = malloc (bytes);
  if (bytes) {
    if (res) memset (res, 0, bytes);
    else die ("out-of-memory allocating %zu bytes", bytes);
  }
  inc_allocated_bytes (bytes);
  return res;
}

static void deallocate (void * ptr, size_t bytes) {
  dec_allocated_bytes (bytes);
  free (ptr);
}

static void * reallocate (void * ptr, size_t old_bytes, size_t new_bytes) {
  dec_allocated_bytes (old_bytes);
  void * res = realloc (ptr, new_bytes);
  if (res) {
    if (new_bytes > old_bytes)
      memset (res + old_bytes, 0, new_bytes - old_bytes);
  } else die ("out-of-memory reallocating %zu bytes", new_bytes);
  inc_allocated_bytes (new_bytes);
  return res;
}

/*------------------------------------------------------------------------*/

// 'polymorphic' array management routine maintain a count of allocated memory

#define ALLOCATE(P,N) \
do { \
  size_t BYTES = (N) * sizeof *(P); \
  (P) = allocate (BYTES); \
} while (0)

#define DEALLOCATE(P,N) \
do { \
  size_t BYTES = (N) * sizeof *(P); \
  deallocate ((P), BYTES); \
} while (0)

#define REALLOCATE(P,O,N) \
do { \
  size_t OLD_BYTES = (O) * sizeof *(P); \
  size_t NEW_BYTES = (N) * sizeof *(P); \
  (P) = reallocate ((P), OLD_BYTES, NEW_BYTES); \
} while (0)

#define SWAP(TYPE,A,B) \
do { \
  TYPE TMP = (A); \
  (A) = (B); \
  (B) = TMP; \
} while (0)
/*------------------------------------------------------------------------*/

// Random odd 64 numbers for constructing hash functions.

static uint64_t nonces[32];
static const size_t num_nonces = sizeof nonces / sizeof nonces[0];

static uint16_t rand16 () {
  int tmp = rand ();
  assert (tmp >= 0);
  uint16_t res = tmp ^ (tmp >> 16);
  return res;
}

static uint64_t rand64 () {
  uint64_t res = 0;
  for (unsigned i = 0; i < 64; i += 16)
    res |= rand16 () << i;
  return res;
}

static void init_nonces () {
  srand (42);
  for (size_t i = 0; i < num_nonces; i++)
    nonces[i] = rand64 () | 1;
}

/*------------------------------------------------------------------------*/

// String functions.

static char * duplicate_string (const char * str) {
  if (!str) return 0;
  size_t bytes = strlen (str) + 1;
  char * res = allocate (bytes);
  assert(res);
  strcpy (res, str);
  return res;
}

static void deallocate_string (char * str) {
  DEALLOCATE (str, strlen (str) + 1);
}

static uint64_t hash_string (const char * str) {
  uint64_t res = 0;
  size_t i = 0;
  for (const char * p = str; *p; p++) {
    res += *p;
    res *= nonces[i++];
    if (i == num_nonces) i = 0;
  }
  return res;
}

/*------------------------------------------------------------------------*/

// Variables are kept canonical using a global hash table similar to a
// symbol table in an ordinary compiler.

typedef struct Variable Variable;

struct Variable {
  char * name;			// variable name
  uint64_t hash;		// saved hash value
  Variable * next;		// hash collision chain link
  char * pre;
  int fliped;
};


static uint64_t num_variables;
static uint64_t size_variables;
static Variable ** variable_table;

static void enlarge_variables () {
  uint64_t new_size_variables = size_variables ? 2*size_variables : 1;
  Variable ** new_variable_table;
  ALLOCATE (new_variable_table, new_size_variables);
  for (uint64_t i = 0; i < size_variables; i++) {
    for (Variable * v = variable_table[i], * n; v; v = n) {
      uint64_t h = v->hash & (new_size_variables - 1);
      n = v->next;
      v->next = new_variable_table[h];
      new_variable_table[h] = v;
    }
  }
  DEALLOCATE (variable_table, size_variables);
  variable_table = new_variable_table;
  size_variables = new_size_variables;
}



static int is_valid_letter (char ch) {
  if (ch == '_') return 1;
  if ('a' <= ch && ch <= 'z') return 1;
  if ('A' <= ch && ch <= 'Z') return 1;
  return 0;
}



#ifndef NDEBUG

static int is_valid_variable_letter (char ch) {
  if (ch == '_') return 1;
  if ('0' <= ch && ch <= '9') return 1;
  if ('a' <= ch && ch <= 'z') return 1;
  if ('A' <= ch && ch <= 'Z') return 1;
  return 0;
}

static int is_valid_variable_name (const char * name) {
  if (!name) return 0;
  for (const char * p = name; *p; p++)
    if (!is_valid_variable_letter (*p)) return 0;
  return 1;
}

#endif

static Variable * new_variable (const char * name, int flipped) {
  assert (is_valid_variable_name (name));
  if (num_variables == size_variables) enlarge_variables ();
  char * pre = flipped ? "n" : "l";

  const uint64_t hash = hash_string (name);

  const uint64_t h = hash & (size_variables - 1);
  Variable ** p, * res;
  for (p = variable_table + h;
       (res = *p) && (strcmp (res->name, name) || strcmp(res->pre, pre));
       p = &res->next)
    ;
  if (!res) {
    *p = res = allocate (sizeof *res);
    res->name = duplicate_string (name);
    res->hash = hash;
    res->next = 0;
    res->pre = pre;
    res->fliped = 0;
    num_variables++;
  }

  return res;
}


static void print_variable (Variable * v, FILE * file) {
  fputs (v->pre, file);
  fputs (v->name, file);
}

static int is_flipped_variable_of (Variable * v, Variable * w) {
  if(v == w ) return 0;
  if(strcmp(v->name, w->name)!=0) {
    return 0;
  }
  else if (strcmp(v->pre, w->pre)!=0) {
    return 1;
  }
  return 0;
}

static int is_flipped_variable (Variable * v) {
  return strcmp (v->pre, "n") == 0;
}


static Variable * flip_variable (Variable * v) {
  int sign = 1;
  if(is_flipped_variable(v)) sign = 0;
  Variable * v1 = new_variable(v->name, sign);
  return v1;
}


static int cmp_variable_val (Variable * a, Variable * b) {
  if (a == b) return 0;

  if (strlen(a->name)> strlen(b->name)) return -1;
  else if (strlen(a->name)< strlen(b->name)) return 1;
  else return -1*strcmp(a->name, b->name);
}


static int cmp_variable (Variable * a, Variable * b) {
  if (a == b) return 0;

  int cmp_name = cmp_variable_val(a, b);
  if (cmp_name) return cmp_name;
  else return -1*strcmp(a->pre, b->pre);
}


static void deallocate_variables () {
  for (uint64_t i = 0; i < size_variables; i++) {
    for (Variable * v = variable_table[i], * n; v; v = n) {
      n = v->next;
      deallocate_string (v->name);
      deallocate (v, sizeof *v);
    }
  }
  DEALLOCATE (variable_table, size_variables);
}

/*----------------------------------------------------------------------------*/
typedef struct Term Term;

struct Term {
  Variable * variable;  // variable name
  Term * rest;     // single linked list
  uint64_t ref;	   // reference counter
  uint64_t hash;   // saved hash value
  Term * next;	   // hash collision chain link
};

// Hash table to canonical store terms.
static uint64_t size_terms;
static uint64_t current_terms;
static Term ** term_table;


// compute hashvalues for terms
static uint64_t
compute_hash_term (Variable * variable, Term * rest) {
  assert (variable);
  uint64_t res = rest ? rest->hash : 0;
  res *= nonces[0];
  res += variable->hash;
  res *= nonces[1];
  return res;
}


// enlarge hash table
static void enlarge_terms () {
  uint64_t new_size_terms = size_terms ? 2*size_terms : 1;
  Term ** new_term_table;
  ALLOCATE (new_term_table, new_size_terms);
  for (uint64_t i = 0; i < size_terms; i++) {
    for (Term * m = term_table[i], * n; m; m = n) {
      uint64_t h = m->hash & (new_size_terms - 1);
      n = m->next;
      m->next = new_term_table[h];
      new_term_table[h] = m;
    }
  }
  DEALLOCATE (term_table, size_terms);
  term_table = new_term_table;
  size_terms = new_size_terms;
}


// copy a term
static Term * copy_term (Term * m) {
  if (!m) return m;
  assert(m);
  assert (m->ref > 0);

  m->ref++;
  assert (m->ref);
  return m;
}


// Then searching and adding a term extended by one power.
static Term *
new_term (Variable * variable, Term * rest) {
  if (rest) assert (cmp_variable(variable, rest->variable) > 0);
  if (current_terms == size_terms) enlarge_terms ();
  const uint64_t hash = compute_hash_term (variable, rest);
  const uint64_t h = hash & (size_terms - 1);

  Term * res;
  for (res = term_table[h];
       res && (res->variable != variable || res->rest != rest);
       res = res->next)
       ;

  if (res) {
    res->ref++; //here we extend that we found monomomial once more
  } else {
    res = allocate (sizeof *res);
    res->variable = variable;
    res->rest = copy_term (rest);
    res->ref = 1;
    res->hash = hash;
    res->next = term_table[h];
    term_table[h] = res;
    current_terms++;

  }
  return res;
}

// Decrements the reference count of a term, and actually deletes a
// term if its reference count goes to zero.  In this case it also
// removes it from the hash table and applies the same procedure to the
// suffix 'rest'.
static void deallocate_term (Term * m) {
  while (m) {
    assert (m->ref > 0);
    if (--m->ref > 0) break; //we still need m
    Term * rest = m->rest;
    const uint64_t h = m->hash & (size_terms - 1);
    Term ** p = term_table + h, * n;
    while ((n = *p) != m) assert (n), p = &n->next;
    *p = n->next;
    assert (current_terms);
    current_terms--;
    deallocate (m, sizeof *m);
    m = rest;
  }
}

// deallocate all terms
static void deallocate_terms () {
  for (uint64_t i = 0; i < size_terms; i++) {
    for (Term * m = term_table[i], *n; m; m = n) {
      n = m->next;
      assert (current_terms);
      current_terms--;
      deallocate (m, sizeof *m);
    }
  }
  DEALLOCATE (term_table, size_terms);
}

// print term
static void print_term (Term * m, FILE * file) {
  Term * tmp = m;
  while(tmp){
    print_variable(tmp->variable, file);
    tmp = tmp->rest;
    if(tmp) fputc_unlocked('*', file);
  }
}

// compare terms
static int cmp_terms(Term * t1, Term * t2){
  Term * tmp1 = t1;
  Term * tmp2 = t2;

  while(tmp1 && tmp2){
    int cmp_var = cmp_variable(tmp1->variable, tmp2->variable);
    if (cmp_var > 0) return 1;
    else if (cmp_var < 0) return -1;

    tmp1 = tmp1->rest;
    tmp2 = tmp2->rest;
  }
  if (tmp1) return 1;
  else if (tmp2) return -1;

  return 0;
}


/*------------------------------------------------------------------------*/
// Define a stack for sorting variables in terms
static size_t size_vstack;
static size_t num_vstack = 0;
static Variable** vstack;

// make buffer bigger
static void enlarge_vstack () {
  size_t new_size_vstack = size_vstack ? 2*size_vstack : 1;
  REALLOCATE (vstack, size_vstack, new_size_vstack);
  size_vstack = new_size_vstack;
}

// push char on buffer
static void push_vstack (Variable*v) {
  if (size_vstack == num_vstack) enlarge_vstack ();
  vstack[num_vstack++] = v;
}

// set size of buffer to zero
static void clear_vstack () { num_vstack = 0; }

// deallocate whole buffer
static void deallocate_vstack () { DEALLOCATE (vstack, size_vstack); }

// Generate a term from the variables on the stack
static Term * build_term_from_stack () {
  Term * res = 0;

  for(int j = num_vstack-1; j>=0; j--){
    Term * t = new_term(vstack[j], res);
    deallocate_term(res);
    res = t;
  }

  clear_vstack();
  return res;
}

/*------------------------------------------------------------------------*/

typedef struct Monomial Monomial;

struct Monomial {
  mpz_t coeff;
  Term * term;
};

// generate new monomial
static Monomial * new_monomial (mpz_t c, Term * m) {
 Monomial * res = allocate (sizeof *res);
 mpz_init(res->coeff);
 mpz_set(res->coeff, c);

 if (mpz_cmp_si(c, 0) == 0) res->term = 0;
 else  res->term = m;

 return res;
}


// copy a monomial
static Monomial * copy_monomial (Monomial * a) {
  assert(a);
  Monomial * t = new_monomial(a->coeff, copy_term(a->term));
  assert(t);
  return t;
}

// deallocate a monomial
static void deallocate_monomial(Monomial * t){
  mpz_clear(t->coeff);
  deallocate_term(t->term);
  deallocate(t, sizeof *t);
}

// print a monomial
static void print_monomial (Monomial * t, FILE * file, int lm){
  int sign = mpz_cmp_si(t->coeff, 0);
  if (!sign) return;
  else if (!lm && sign > 0) fputc_unlocked('+',file);

  if(!t->term)  mpz_out_str(file, 10, t->coeff);
  else if (mpz_cmp_si(t->coeff, -1) == 0) fputc_unlocked('-',file);
  else if (mpz_cmp_si(t->coeff, 1) != 0) mpz_out_str(file, 10, t->coeff);

  if (t->term) {
    if(mpz_cmp_ui(t->coeff, 1) != 0 && mpz_cmp_si(t->coeff, -1) != 0) fputc_unlocked('*',file);
    print_term(t->term, file);
  }
}

// compare two monomials
/*static int cmp_monomials(Monomial * t1, Monomial * t2){
  int term = cmp_terms(t1->term, t2->term);
  if (term) return term;

  return mpz_cmp(t1->coeff, t2->coeff);
}*/


/*------------------------------------------------------------------------*/
typedef struct Polynomial Polynomial;

struct Polynomial {
  int idx;
  Monomial * lm;
  Polynomial * rest; //single linked list
};


// generate new polynomial
static Polynomial *
new_polynomial (Monomial * lm, Polynomial * rest) {
  if (rest) assert (cmp_terms(lm->term, rest->lm->term) >= 0);

  Polynomial * res = allocate (sizeof *res);
  res->lm = lm;
  res->rest = rest;

  return res;
}

static int is_one_poly (Polynomial* p){
  if(!p) return 0;
  if(p->rest) return 0;
  if(p->lm->term) return 0;
  if(mpz_cmp_si(p->lm->coeff, 1) != 0) return 0;
  return 1;
}


// copy a poly
static Polynomial * copy_poly (Polynomial*p){
  if(!p) return 0;
  Polynomial * tmp = p;
  Polynomial * p1 = new_polynomial(copy_monomial(tmp->lm), 0);
  Polynomial * out = p1;
  tmp = tmp->rest;

  while(tmp){
    Polynomial * res = new_polynomial(copy_monomial(tmp->lm), 0);
    p1->rest = res;
    tmp = tmp->rest;
    p1 = p1->rest;
  }
  assert(out);
  return out;
}

// deallocate a term
static void deallocate_polynomial (Polynomial * p) {
  while (p){
    deallocate_monomial(p->lm);
    Polynomial * rest = p->rest;
    deallocate (p, sizeof *p);
    p = rest;
  }
}

// print a polynomial
static void print_polynomial (Polynomial * p, FILE * file){
  if(!p) fputs("0",file);
  else if(mpz_cmp_si(p->lm->coeff, 0) == 0) fputs("0", file);
  else {
    Polynomial * tmp = p;
    int lm = 1;
    while(tmp){
      print_monomial(tmp->lm, file, lm);
      lm = 0;
      tmp=tmp->rest;
    }
  }
}

/*static int cmp_polynomials (Polynomial * tmp1, Polynomial * tmp2){

  while(tmp1 && tmp2){
    int cmp = cmp_monomials(tmp1->lm, tmp2->lm);
    if(cmp) return cmp;

    tmp1= tmp1->rest;
    tmp2= tmp2->rest;
  }
  if(tmp1) return 1;
  else if(tmp2) return -1;
  return 0;
}*/


/*------------------------------------------------------------------------*/
static size_t size_mstack;
static size_t num_mstack = 0;
static Monomial ** mstack;

// make buffer bigger
static void enlarge_mstack () {
  size_t new_size_mstack = size_mstack ? 2*size_mstack : 1;
  REALLOCATE (mstack, size_mstack, new_size_mstack);
  size_mstack = new_size_mstack;
}

// push char on buffer
static void push_mstack (Monomial * t) {
  assert(t);
  if (mpz_cmp_si(t->coeff, 0) == 0) {
    deallocate_monomial(t);
    return;
  }
  if (size_mstack == num_mstack) enlarge_mstack ();

  if (num_mstack == 0) mstack[num_mstack++] = t;

  else if (cmp_terms(mstack[num_mstack-1]->term, t->term) > 0) mstack[num_mstack++] = t;

  else
  {
    assert(num_mstack > 0);
    int i = num_mstack - 1;
    int cmp = -1;
    Monomial * tmp = 0;

    while (i >= 0) {
      tmp = mstack[i];
      cmp = cmp_terms(tmp->term, t->term);
      if (cmp >= 0) break;
      i--;
    }

    assert(i >= 0);

    if (cmp == 0) {
      mpz_add(tmp->coeff, tmp->coeff, t->coeff);
      deallocate_monomial(t);

      if(mpz_cmp_si(tmp->coeff, 0) == 0){
        for(unsigned j = i; j < num_mstack-1; j++)
        mstack[j] = mstack[j+1];

        num_mstack--;
        deallocate_monomial(tmp);
      }
    }
    else {
      for (int j = num_mstack; j > i+1; j--)
        mstack[j] = mstack[j-1];

      mstack[i+1] = t;
      num_mstack++;
    }
  }
}

static void push_mstack_sorted(Monomial * t){
  assert(t);
  if (mpz_cmp_si(t->coeff, 0) == 0) {
    deallocate_monomial(t);
    return;
  }
  if (size_mstack == num_mstack) enlarge_mstack ();

  mstack[num_mstack++] = t;
}


// set size of buffer to zero
static void clear_mstack () { num_mstack = 0; }

// deallocate whole buffer
static void deallocate_mstack () { DEALLOCATE (mstack, size_mstack); }

// build a polynomial from stack
static Polynomial * build_polynomial_from_stack () {
  Polynomial * res = 0;

  for (int i = num_mstack-1; i >= 0; i--) {
    Polynomial * tmp = new_polynomial (mstack[i], res);
    res = tmp;
  }

  clear_mstack();
  return res;
}



/*------------------------------------------------------------------------*/
// Polynomial operations: negation, addition, multiplication

// negate all monomials in a polynomial
static Polynomial * negate_poly(Polynomial * p1){
  Polynomial * res = p1;
  while(res){
    mpz_mul_si(res->lm->coeff, res->lm->coeff, -1);
    res = res->rest;
  }
  return p1;
}


//add one polynomial p2 to p1
static Polynomial * add_poly(Polynomial *p1, Polynomial *p2){
  assert(!num_mstack);

  while(p1 && p2){
    if(cmp_terms(p1->lm->term, p2->lm->term) == 1){
      push_mstack_sorted(copy_monomial(p1->lm));
      p1 = p1->rest;
    }
    else if(cmp_terms(p1->lm->term, p2->lm->term) == -1){
      push_mstack_sorted(copy_monomial(p2->lm));
      p2 = p2->rest;
    }
    else{
      Monomial * m = copy_monomial(p1->lm);
      mpz_add(m->coeff, m->coeff, p2->lm->coeff);

      if(mpz_cmp_si(m->coeff, 0) == 0){
        deallocate_monomial(m);
      } else push_mstack_sorted(m);
      p1 = p1->rest;
      p2 = p2->rest;
    }
  }

  while(p1){
    push_mstack_sorted(copy_monomial(p1->lm));
    p1 = p1->rest;
  }
  while(p2){
    push_mstack_sorted(copy_monomial(p2->lm));
    p2 = p2->rest;
  }

  Polynomial * p = build_polynomial_from_stack();
  assert(!num_mstack);
  return p;
}


// multiply two monomials
static Monomial * multiply_monomial(Monomial * t1, Monomial * t2){
  mpz_t ctmp;
  mpz_init(ctmp);
  mpz_mul(ctmp, t1->coeff, t2->coeff);

  Term * tmp1 = t1->term;
  Term * tmp2 = t2->term;

  while(tmp1 && tmp2){
    if( cmp_variable (tmp1->variable, tmp2->variable) > 0){
      push_vstack(tmp1->variable);
      tmp1 = tmp1->rest;
    } else if ( cmp_variable(tmp1->variable, tmp2->variable) < 0){
      push_vstack(tmp2->variable);
      tmp2 = tmp2->rest;
    } else {
      push_vstack(tmp1->variable);
      tmp1 = tmp1->rest;
      tmp2 = tmp2->rest;
    }
  }

  while(tmp1){
    push_vstack(tmp1->variable);
    tmp1 = tmp1->rest;
  }

  while(tmp2){
    push_vstack(tmp2->variable);
    tmp2 = tmp2->rest;
  }

  Term * m = build_term_from_stack();
  Monomial * mon = new_monomial(ctmp, copy_term(m));
  mpz_clear(ctmp);
  return mon;

}

// multiply two polynomials
static Polynomial * multiply_poly(Polynomial *p1, Polynomial *p2){
  assert(!num_mstack);

  Polynomial * tmp1, * start_over;
  if(cmp_terms(p1->lm->term, p2->lm->term)==1){
    tmp1 = p1;
    start_over = p2;
  } else {
    tmp1 = p2;
    start_over = p1;
  }

  while(tmp1){
    Polynomial * tmp2 = start_over;
    while(tmp2){
      push_mstack(multiply_monomial(tmp1->lm, tmp2->lm));
      tmp2=tmp2->rest;
    }
    tmp1=tmp1->rest;
  }

  Polynomial * p = build_polynomial_from_stack();
  return p;
}
/*------------------------------------------------------------------------*/
// Define special Polynomials

// p = -v +1
static Polynomial * define_neg_poly (Variable * v){
  assert(!num_mstack);
  mpz_t pm_one;
  mpz_init(pm_one);
  mpz_set_si(pm_one, -1);

  Term * t1 = new_term(v, 0);
  Monomial * m1 = new_monomial(pm_one, t1);
  push_mstack_sorted(m1);

  mpz_set_ui(pm_one, 1);
  Monomial * m2 = new_monomial(pm_one, 0);
  push_mstack_sorted(m2);

  Polynomial * p = build_polynomial_from_stack();

  assert(!num_mstack);
  mpz_clear(pm_one);
  return p;
}

// p = -l-f+1;
static Polynomial * generate_flip_poly (Variable * l, Variable * f){

  assert(!num_mstack);
  mpz_t pm_one;
  mpz_init(pm_one);
  mpz_set_si(pm_one, -1);

  Term * t1 = new_term(f, 0);
  Monomial * m1 = new_monomial(pm_one, t1);
  push_mstack_sorted(m1);
  Polynomial * q = build_polynomial_from_stack();

  Term * t2 = new_term(l, 0);
  Monomial * m2 = new_monomial(pm_one, t2);
  push_mstack_sorted(m2);

  mpz_set_ui(pm_one, 1);
  Monomial * m3 = new_monomial(pm_one, 0);
  push_mstack_sorted(m3);

  Polynomial * p = build_polynomial_from_stack();
  if(f->fliped == 0){
    fprintf(pac_file, "= ");
    print_term(t1, pac_file);
    fprintf(pac_file, ", ");
    print_polynomial(p, pac_file);
    fprintf(pac_file, ";\n");
    f->fliped = 1;
    l->fliped = 1;
  }
  Polynomial * res = add_poly(q, p);
  deallocate_polynomial(p);
  deallocate_polynomial(q);

  assert(!num_mstack);
  mpz_clear(pm_one);
  return res;
}
/*------------------------------------------------------------------------*/
// Polynomial reduction
static int term_two_is_contained_in_term_one (Term * m1, Term * m2){
  int found = 0;
  Term * tmp1 = m1;
  Term * tmp2 = m2;

  while(tmp1){
    while(tmp2 && !found){
      if(tmp1->variable == tmp2->variable) found = 1;
      tmp2 = tmp2->rest;
    }
    if(!found) return 0;

    tmp1 = tmp1->rest;
    tmp2 = m2;
    found = 0;
  }
  return 1;
}

// if term m2 is contained in m1 it calculates the remainder.
static Term * term_remainder(Term * m1, Term * m2){
  assert(!num_vstack);
  Term * dividend = m1;
  Term * divisor = m2;
  assert(term_two_is_contained_in_term_one (m2, m1));

  if(m1 == m2) return 0;

  while(divisor){
    Variable * v1 = dividend->variable;
    Variable * v2 = divisor->variable;

    if(!cmp_variable(v1, v2)){
      divisor = divisor->rest;
      dividend = dividend->rest;
    }

    if(cmp_variable(v1, v2) > 0){
      push_vstack(v1);
      dividend = dividend->rest;
    }
  }

  while(dividend){
    push_vstack(dividend->variable);
    dividend = dividend->rest;
  }

  Term * m = build_term_from_stack();
  return m;
}

// we divide polynomial p1 by the lm of p2
static Polynomial * divide_lm(Polynomial * p1, Polynomial * p2){
  assert(!num_mstack);
  mpz_t ctmp;
  mpz_init(ctmp);

  Term * lm = p2->lm->term;
  Polynomial * tmp = p1;

  while(tmp){
    Monomial * lm_tmp = tmp->lm;
    if(term_two_is_contained_in_term_one (lm, lm_tmp->term)){
      Term * m = term_remainder(lm_tmp->term, lm);
      mpz_cdiv_q(ctmp, lm_tmp->coeff, p2->lm->coeff);
      push_mstack_sorted(new_monomial(ctmp, copy_term(m)));
    }
    tmp = tmp->rest;
  }
  mpz_clear(ctmp);
  Polynomial * p = build_polynomial_from_stack();
  return p;
}


static Polynomial * reduce_poly_p1_by_poly_p2 (Polynomial * p1, Polynomial * p2, int del){
  Polynomial * negfactor = divide_lm(p1, p2);
  negfactor = negate_poly(negfactor);
  Polynomial * mult   = multiply_poly(negfactor, p2);
  Polynomial * rem    = add_poly(p1, mult);

  if(!is_one_poly(negfactor)){
      if(del)  fprintf(pac_file, "d ");
      fprintf(pac_file, "*: ");
      print_polynomial(p2, pac_file);        fprintf(pac_file, ", ");
      print_polynomial(negfactor, pac_file); fprintf(pac_file, ", ");
      print_polynomial(mult, pac_file);
      fprintf(pac_file, ";\n");
  }

  if(del) fprintf(pac_file, "d ");
  fprintf(pac_file, "+: ");
  print_polynomial(p1, pac_file);     fprintf(pac_file, ", ");
  print_polynomial(mult, pac_file);   fprintf(pac_file, ", ");
  print_polynomial(rem, pac_file);
  fprintf(pac_file, ";\n");

  deallocate_polynomial(negfactor);
  deallocate_polynomial(mult);

  return rem;
}

// we divide polynomial p1 by the lm of p2
static Polynomial * divide_by_second_term(Polynomial * p1, Polynomial * p2){
  assert(!num_mstack);
  mpz_t ctmp;
  mpz_init(ctmp);
  assert(p2->rest);
  Term * lm = p2->rest->lm->term;
  Polynomial * tmp = p1;

  while(tmp){
    Monomial * lm_tmp = tmp->lm;
    if(term_two_is_contained_in_term_one (lm, lm_tmp->term)){
      Term * m = term_remainder(lm_tmp->term, lm);
      mpz_cdiv_q(ctmp, lm_tmp->coeff, p2->lm->coeff);
      push_mstack_sorted(new_monomial(ctmp, copy_term(m)));
    }
    tmp = tmp->rest;
  }
  mpz_clear(ctmp);
  Polynomial * p = build_polynomial_from_stack();
  return p;
}

static Polynomial * unflip_poly_p1_by_poly_p2 (Polynomial * p1, Polynomial * p2){
  Polynomial * negfactor = divide_by_second_term(p1, p2);
  negfactor = negate_poly(negfactor);
  Polynomial * mult   = multiply_poly(negfactor, p2);
  Polynomial * rem    = add_poly(p1, mult);

  if(!is_one_poly(negfactor)){
      fprintf(pac_file, "d *: ");
      print_polynomial(p2, pac_file);        fprintf(pac_file, ", ");
      print_polynomial(negfactor, pac_file); fprintf(pac_file, ", ");
      print_polynomial(mult, pac_file);
      fprintf(pac_file, ";\n");
  }

  fprintf(pac_file, "d +: ");
  print_polynomial(p1, pac_file);     fprintf(pac_file, ", ");
  print_polynomial(mult, pac_file);   fprintf(pac_file, ", ");
  print_polynomial(rem, pac_file);
  fprintf(pac_file, ";\n");

  deallocate_polynomial(negfactor);
  deallocate_polynomial(mult);

  return rem;
}



/*------------------------------------------------------------------------*/
// Clause part
/*------------------------------------------------------------------------*/
typedef struct Lit Lit;

struct Lit {
  Variable * var;			// variable name
  int neg;
};

static Lit * new_lit (int sign, char * lit, int flip) {
  Variable * var = new_variable(lit, flip);
  Lit * l = allocate (sizeof *l);
  l->var = var;
  l->neg = sign;
  return l;
}

static Lit * copy_lit (Lit * old){
  Lit * l = allocate (sizeof *l);
  l->var = old->var;
  l->neg = old->neg;
  return l;
}

static void print_lit (Lit * l, FILE * file) {
  if(l->neg == 1) fprintf(file, "-");
  print_variable(l->var, file);
}

static void delete_lit (Lit * l) {
  deallocate (l, sizeof *l);
}



/*------------------------------------------------------------------------*/


typedef struct Clause Clause;

struct Clause {
  Lit * lit;
  Clause * next;
  int64_t size;
};

static Clause * new_clause (Lit * lit, Clause * next) {
  Clause * c = allocate (sizeof *c);
  c->lit = lit;
  c->next = next;
  c->size = 0;
  return c;
}

static Clause * copy_clause(Clause*t){
  if(!t) return 0;
  Clause * tmp = t;
  Clause * t1 = new_clause(copy_lit(tmp->lit), 0);
  Clause * out = t1;
  tmp = tmp->next;

  while(tmp){
    Clause * res = new_clause(copy_lit(tmp->lit), 0);
    t1->next = res;
    tmp = tmp->next;
    t1 = t1->next;
  }
  assert(out);
  out->size = t->size;
  return out;
}


static void print_clause (Clause * c, FILE * file) {
  if(!c) return;
  while(c){
    print_lit(c->lit, file);
    fputc(' ', file);
    c=c->next;
  }
}
/*
static int is_empty_clause (Clause * c) {
  if(!c) return 1;
  else if(!c->lit) return 1;
  return 0;
}*/

static int is_unit_clause (Clause * c) {
  if(!c) return 0;
  if(c->lit && !c->next) return 1;
  return 0;
}


static int contain_same_variables (Clause * c1, Clause * c2) {
/*  msg("compare with");
  fprintf(stdout, "clause 1: ");
  print_clause(c1, stdout);
  fprintf(stdout, "\nclause 2: ");
  print_clause(c2, stdout);
  fprintf(stdout, "\n\n");*/
  while(c1 && c2){
    if(!cmp_variable_val(c1->lit->var, c2->lit->var)){
      c1 = c1->next;
      c2 = c2->next;
    }
    else return 0;
  }
  if(c1 || c2) return 0;
  return 1;
}

// 1 if lit , 0 if not found - 1 if opposit found
static int find_lit_in_clause(Lit * lit, Clause * cl){
  if(!cl) return 0;
  while(cl){
    if(lit->var == cl->lit->var) return 1;
    if(is_flipped_variable_of(lit->var, cl->lit->var)) return -1;

    cl= cl->next;
  }
  return 0;
}

static void flip_var_in_clause(Clause * cl, Variable * old_v, Variable * new_v){
  if(!cl) return;
  while(cl){
    if(!cmp_variable(old_v, cl->lit->var)){
      cl->lit->var = new_v;
      cl->lit->neg = (cl->lit->neg == 0);
      return;
    }
    cl= cl->next;
  }
}

static void deallocate_clause (Clause * c) {
  Clause * tmp;
  while(c){
    tmp = c->next;
    delete_lit(c->lit);
    deallocate (c, sizeof *c);
    c=tmp;

  }
}

/*------------------------------------------------------------------------*/

static int64_t num_lits;
static int64_t size_lits;
static Lit ** lits;

static void enlarge_lits () {
  size_t new_size_lits = size_lits ? 2*size_lits : 1;
  REALLOCATE (lits, size_lits, new_size_lits);
  size_lits = new_size_lits;
}

static void push_lit (Lit * l) {
  if (size_lits == num_lits) enlarge_lits ();
  int64_t i = 0;
  Lit * tmp = 0;
  if(!num_lits){
    lits[num_lits++] = l;
    return;
  }

  if(cmp_variable(lits[num_lits-1]->var, l->var) > 0) {
    lits[num_lits++] = l;
    return;
  }

  while(i < num_lits){
    tmp = lits[i];
    if(cmp_variable(tmp->var, l->var) <= 0 ) break;
    i++;
  }

  if(tmp->var == l->var){
    if(tmp->neg == l->neg) {
      delete_lit(l);
      return;
    }
    else{
      Lit * del = lits[i];
      for(int64_t j=i; j<num_lits-1; j++){
        lits[j] = lits[j+1];
      }
      delete_lit(del);
      delete_lit(l);
      num_lits--;
      return;
    }
  }

  for(int64_t j=num_lits; j>i; j--){
    lits[j]=lits[j-1];
  }
  lits[i] = l;
  num_lits++;
}

static void push_lit_sorted (Lit * l) {
  if (size_lits == num_lits) enlarge_lits ();
  lits[num_lits++] = l;
}


static void clear_lits () { num_lits = 0; }

static void deallocate_lits () { DEALLOCATE (lits, size_lits); }

static Clause * build_clause(){
  Clause * next = 0;
  for(int64_t i = num_lits-1; i>=0; i--){
    Clause * c = new_clause(lits[i], next);
    next = c;
  }
  if(num_lits) next->size = num_lits;
  clear_lits();
  return next;
}

/*------------------------------------------------------------------------*/
static Clause * divide_clause(Clause * c1, Clause * c2){
  while(c1 && c2){
    int cmp = cmp_variable(c1->lit->var, c2->lit->var);
    if(cmp < 0) c2 = c2->next;
    else if (cmp > 0) {
      push_lit_sorted(copy_lit(c1->lit));
      c1 = c1->next;
    }
    else {
      c1 = c1->next;
      c2 = c2->next;
    }
  }

  while(c1){
    push_lit_sorted(copy_lit(c1->lit));
    c1 = c1->next;
  }

  Clause * c = build_clause();
  return c;
}

static Clause * merge_two_clauses(Clause * c1, Clause * c2){
  while(c1 && c2){
    int cmp = cmp_variable(c1->lit->var, c2->lit->var);
    if(cmp > 0) {
      push_lit_sorted(copy_lit(c2->lit));
      c2 = c2->next;
    }
    else if (cmp < 0) {
      push_lit_sorted(copy_lit(c1->lit));
      c1 = c1->next;
    }
    else {
      if (c1->lit->neg == c2->lit->neg) push_lit(copy_lit(c1->lit));
      c1 = c1->next;
      c2 = c2->next;
    }
  }

  while(c1){
    push_lit(copy_lit(c1->lit));
    c1 = c1->next;
  }
  while(c2){
    push_lit(copy_lit(c2->lit));
    c2 = c2->next;
  }

  Clause * c = build_clause();
  return c;
}

/*------------------------------------------------------------------------*/

static Polynomial * pos_poly_from_clause (Clause * c){
  mpz_t one;
  mpz_init(one);
  mpz_set_ui(one, 1);

  while(c){
    if(c->lit->neg == 1 ) push_vstack(c->lit->var);
    c = c->next;
  }
  Polynomial * pos = 0;
  if(num_vstack) {
    Term * t = build_term_from_stack();
    Monomial * m = new_monomial(one, t);
    pos = new_polynomial(m, 0);
  }
  mpz_clear(one);
  return pos;

}

static Polynomial * neg_poly_from_clause (Clause * c){
  Polynomial * neg = 0;
  while(c){
    if(c->lit->neg == 0){
      Polynomial * fac = define_neg_poly(c->lit->var);
      if(neg){
        Polynomial * mul = multiply_poly(neg, fac);
        deallocate_polynomial(fac);
        deallocate_polynomial(neg);
        neg = mul;
      } else neg = fac;
    }
    c = c->next;
  }
  return neg;
}


static Polynomial * poly_from_clause (Clause * clause){

  Polynomial * pos = pos_poly_from_clause(clause);
  Polynomial * neg = neg_poly_from_clause(clause);

  Polynomial * res = 0;
  if(pos && neg){
    res = multiply_poly(pos, neg);
    deallocate_polynomial(pos);
    deallocate_polynomial(neg);
  } else if (pos)  res = pos;
    else res = neg;

  return res;
}
/*------------------------------------------------------------------------*/

typedef struct Trace Trace;

struct Trace {
  int cl;			// variable name
  Trace * next;
};

static Trace * new_trace (int cl, Trace * next) {
  Trace * t = allocate (sizeof *t);
  t->cl = cl;
  t->next = next;
  return t;
}


// copy a poly
static Trace * copy_trace(Trace*t){
  if(!t) return 0;
  Trace * tmp = t;
  Trace * t1 = new_trace(tmp->cl, 0);
  Trace * out = t1;
  tmp = tmp->next;

  while(tmp){
    Trace * res = new_trace(tmp->cl, 0);
    t1->next = res;
    tmp = tmp->next;
    t1 = t1->next;
  }
  assert(out);
  return out;
}

static void print_trace (Trace * t, FILE * file) {
  while(t){
    fprintf(file, "%i ", t->cl);
    t=t->next;
  }
}

static void delete_trace (Trace * t) {
  Trace * tmp;
  while(t){
    tmp = t->next;
    deallocate (t, sizeof *t);
    t=tmp;

  }
}

/*------------------------------------------------------------------------*/
static int64_t num_traces;
static int64_t size_traces;
static uint64_t * traces;

static void enlarge_traces () {
  size_t new_size_traces = size_traces ? 2*size_traces : 1;
  REALLOCATE (traces, size_traces, new_size_traces);
  size_traces = new_size_traces;
}

static void push_trace (uint64_t idx) {
  assert(idx > 0);
  if (size_traces == num_traces) enlarge_traces ();
  int64_t i = 0;
  uint64_t tmp = 0;
  if(!num_traces){
    traces[num_traces++] = idx;
    return;
  }

  if(idx > traces[num_traces-1]){
    traces[num_traces++] = idx;
    return;
  }

  while(i < num_traces){
    tmp = traces[i];
    if(tmp >= idx) break;
    i++;
  }

  if(tmp == idx) return;

  for(int64_t j = num_traces; j>i; j--){
    traces[j] = traces[j-1];
  }
  traces[i] = idx;
  num_traces++;


}

static void clear_traces () { num_traces = 0; }

static void deallocate_traces () { DEALLOCATE (traces, size_traces); }

static Trace * build_trace(){
  Trace * next = 0;
  for(int64_t i = num_traces-1; i>=0; i--){
    Trace * t = new_trace(traces[i], next);
    if(next) assert(next->cl > t->cl);
    next = t;
  }
  clear_traces();
  return next;
}



static Trace * add_to_trace(Trace * whole, int cl){
  Trace * insert = new_trace(cl, 0);
  Trace * t = whole;
  while(t->next) t = t->next;
  t->next = insert;
  return whole;
}

static Trace * delete_from_trace(Trace * whole, int cl){
  Trace * t = whole;
  Trace * tmp;

  if(t->cl == cl){
    tmp = t->next;
    deallocate (t, sizeof *t);
    return tmp;
  }
  else {
    while(t->next){
      if(t->next->cl == cl){
        tmp = t->next;
        t->next = tmp->next;
        deallocate (tmp, sizeof *tmp);
        return whole;
      }
      t = t->next;
    }
  }
   return whole;
}


/*------------------------------------------------------------------------*/
// Variables are kept canonical using a global hash table similar to a
// symbol table in an ordinary compiler.
static int64_t rule_count;
typedef struct Rule Rule;

struct Rule {
  int idx;
  Clause * cl;
  Trace * tr;
  Polynomial * pl;
};

static Rule * new_rule (int idx, Clause * cl, Trace * tr, Polynomial * pl) {
  Rule * r = allocate (sizeof *r);
  if(idx) r->idx = idx;
  else r->idx = rule_count;
  r->cl = cl;
  r->tr = tr;
  r->pl = pl;
  rule_count++;
  return r;
}


static void print_rule (Rule * r, FILE * file) {
  fprintf(file, "rule: %i \n", r->idx);
  fprintf(file, "clause: ");
  print_clause(r->cl, file);
  fprintf(file, "\ntrace: ");
  print_trace(r->tr, file);
  fprintf(file, "\npolynomial: ");
  print_polynomial(r->pl, file);
  fprintf(file, "\n\n");
}

static void delete_rule (Rule * r) {
  deallocate_clause(r->cl);
  delete_trace(r->tr);
  deallocate_polynomial(r->pl);
  deallocate (r, sizeof *r);
}


static void set_poly_in_rule (Rule * r) {
  Clause * c = r->cl;
  if(!c) return;
  Polynomial * res = poly_from_clause(c);
  r->pl = res;
}

/*------------------------------------------------------------------------*/
static Rule * bad_target;
/*------------------------------------------------------------------------*/

static size_t size_rules;
static size_t num_rules;
static Rule ** rules;

static void enlarge_rules () {
  size_t new_size_rules = size_rules ? 2*size_rules : 1;
  REALLOCATE (rules, size_rules, new_size_rules);
  size_rules = new_size_rules;
}

static void push_rule (Rule * r) {
  if (size_rules == num_rules) enlarge_rules ();
  rules[num_rules++] = r;
}

/*
static void print_all_rules(){
  for(size_t i = 0; i < num_rules; i++){
    print_rule(rules[i], stdout);
  }
}

static void print_all_non_trace_rules(){
  for(size_t i = 0; i < start_traces; i++){
    print_rule(rules[i], stdout);
  }
}*/

static Rule * find_rule (int64_t start, int64_t end, int64_t goal){
  Rule * ende = rules[end];
  if (ende->idx == goal) return ende;
  assert(start <= end);

  if (start == end){
    Rule * r = rules[start];
    assert(r);
    if (r->idx != goal) {
      return 0;
    }
    else return r;
  }
  int64_t middle = (start + end)/2;
  Rule * mid = rules[middle];
  if(mid->idx == goal) return mid;
  assert(mid);

  if (mid->idx >= goal) return find_rule(start, middle, goal);
  else return find_rule(middle+1, end, goal);
  /*for(int i = 0; i<num_rules; i++){
    Rule * r = rules[i];
    if(r->idx == goal) return r;
  }*/
  return 0;


}

static int find_rule_idx (int64_t start, int64_t end, int64_t goal){
  assert(start <= end);
  if (start == end){
    Rule * r = rules[start];
    assert(r);
    if (r->idx != goal) return -1;
    else return start;
  }
  int64_t middle = (start + end)/2;
  Rule * mid = rules[middle];
  assert(mid);

  if (mid->idx >= goal) return find_rule_idx(start, middle, goal);
  else return find_rule_idx(middle+1, end, goal);
}

static void delete_from_rules(Rule * r){

  int res = find_rule_idx (0, num_rules-1, r->idx);
  if (res == -1 ) return;

  for (unsigned i = res; i < num_rules-1; i++){
    rules[i] = rules[i+1];
  }
  num_rules--;
  delete_rule(r);
}

static void clean_up_rules(){

  for (unsigned i = end_traces; i < num_rules; i++){
    Rule * r = rules[i];
    delete_rule(r);
  }

  num_rules = end_traces;

}

static int is_bad_target_clause (Rule * r ) {
  if(r->cl->size > 1) return 0;
  Variable * bad = bad_target->cl->lit->var;
  Variable * cmp = r->cl->lit->var;
  if(is_flipped_variable_of(bad, cmp)) return 1;
  return 0;

}

static void deallocate_rules () {
  for(uint64_t i = 0; i<num_rules; i++){
    delete_rule(rules[i]);
  }

  DEALLOCATE (rules, size_rules); }

/*#############################################################################*/

static size_t size_units;
static size_t num_units;
static Rule ** units;

static void enlarge_units () {
  size_t new_size_units = size_units ? 2*size_units : 1;
  REALLOCATE (units, size_units, new_size_units);
  size_units = new_size_units;
}

static void push_unit (Rule * r) {
  if (size_units == num_units) enlarge_units ();
  units[num_units++] = r;
}

static void deallocate_units () {
  DEALLOCATE (units, size_units); }
  /*-------------------------------------------------------------------------*/
  static size_t size_pairs;
  static size_t num_pairs;
  static Rule ** pairs;

  static void enlarge_pairs () {
    size_t new_size_pairs = size_pairs ? 2*size_pairs : 1;
    REALLOCATE (pairs, size_pairs, new_size_pairs);
    size_pairs = new_size_pairs;
  }

  static void push_pair (Rule * r) {
    if (size_pairs == num_pairs) enlarge_pairs ();
    pairs[num_pairs++] = r;
    //print_rule(r, stdout);
  }

  static void delete_from_pairs (unsigned idx) {
    for(unsigned i = idx; i<num_pairs-1; i++){
      pairs[i]=pairs[i+1];
    }
    num_pairs--;
  }

  static void deallocate_pairs () {
    DEALLOCATE (pairs, size_pairs); }
/*-------------------------------------------------------------------------*/

static Polynomial * flip_lit_in_rule(Rule * r, Variable * v, Variable * w, Polynomial * p){
  Polynomial * rem = 0;
  if(r->cl) flip_var_in_clause(r->cl, v, w);
  if(r->pl) rem = reduce_poly_p1_by_poly_p2 (r->pl, p, 0);

  return rem;
}


static void bit_flip_rule(Rule * r){
  Clause * cl = r->cl;
  while (cl){
    Lit * l = cl->lit;
    if(!l->neg){
      Variable * v = l->var;
      Variable * w = flip_variable(v);
      Polynomial * flip_poly = generate_flip_poly(v, w);
      Polynomial * p = flip_lit_in_rule(r, v, w, flip_poly);
      if(p && r->pl) {
        deallocate_polynomial(r->pl);
        r->pl = p;
      }
      deallocate_polynomial(flip_poly);
    }
    cl = cl->next;
  }
}


/*------------------------------------------------------------------------*/
static Variable * resolv_variable;

static Rule * do_resolution_step(Rule * r1, Rule * r2){
  int del = 0;
  Rule * end = rules[end_traces-1];
  if(r1->idx > end->idx && r2->idx > end->idx) del = 1;

  Clause * c1 = divide_clause(r1->cl, r2->cl);
  Clause * c2 = divide_clause(r2->cl, r1->cl);
  Polynomial * f1 = copy_poly(r1->pl);
  Polynomial * f2 = copy_poly(r2->pl);

  Clause * merge_clause = merge_two_clauses(r1->cl, r2->cl);

  if(c1){
    Polynomial * p1 = poly_from_clause(c1);
    Polynomial * mul = multiply_poly(p1, f2);
    if(del)  fprintf(pac_file, "d ");
    fprintf(pac_file, "*: ");
    print_polynomial(f2, pac_file); fprintf( pac_file, ", ");
    print_polynomial(p1, pac_file); fprintf( pac_file, ", ");
    print_polynomial(mul, pac_file); fprintf( pac_file, ";\n");

    deallocate_polynomial(p1);
    deallocate_polynomial(f2);
    f2 = mul;
  }
  if(c2){
    Polynomial * p2 = poly_from_clause(c2);
    if(del)  fprintf(pac_file, "d ");
    fprintf(pac_file, "*: ");
    print_polynomial(f1,  pac_file); fprintf( pac_file, ", ");
    print_polynomial(p2,  pac_file); fprintf( pac_file, ", ");
    Polynomial * mul = multiply_poly(p2, f1);
    print_polynomial(mul,  pac_file); fprintf( pac_file, ";\n");


    deallocate_polynomial(p2);
    deallocate_polynomial(f1);
    f1 = mul;
  }
  deallocate_clause(c1);
  deallocate_clause(c2);


  Polynomial * add = add_poly(f1, f2);
  if(del)  fprintf(pac_file, "d ");
  fprintf(pac_file, "+: ");
  print_polynomial(f1, pac_file); fprintf(pac_file, ", ");
  print_polynomial(f2, pac_file); fprintf(pac_file, ", ");
  print_polynomial(add, pac_file); fprintf(pac_file, ";\n");


  deallocate_polynomial (f1);
  deallocate_polynomial (f2);

  Rule * r = new_rule(0, merge_clause, 0, add);
  return r;
}


static Rule * do_resolution_step_flipped(Rule * r1, Rule * r2, Variable * v){

  Rule * not_flipped = r1;
  Rule * flipped = r2;

  if(!v) {
    v = resolv_variable;
    resolv_variable = 0;
  }

  // get returned resolution var in r1->cl - check if it is flipped to
  // determine which rules to pass on.
  if(!v) die("no resolving variable found");

  Variable * w = flip_variable(v);
  if(is_flipped_variable(v)) {
    not_flipped = r2;
    flipped = r1;
    Variable * tmp = v;
    v = w;
    w = tmp;
  }


  Polynomial * flip_poly = generate_flip_poly(v, w);
  Polynomial * res = reduce_poly_p1_by_poly_p2 (not_flipped->pl, flip_poly, 0);
  deallocate_polynomial(flip_poly);

  Clause * cl = copy_clause(not_flipped->cl);
  flip_var_in_clause(cl, v, w);

  Rule * tmp = new_rule(0, cl, 0, res);
  Rule * rem = do_resolution_step(flipped, tmp);
  delete_rule(tmp);

  return rem;
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void unit_propagation_step(Rule * unit){

  for(uint64_t i = 0; i<start_traces; i++){
    Rule * r = rules[i];
    if(r->idx == (int)bad_clause) continue;
    if(r->idx == unit->idx) continue;
    if(!r->cl) continue;
    int lit_in_clause = find_lit_in_clause(unit->cl->lit, r->cl);
    if (lit_in_clause == -1 ){
      Rule * tmp = do_resolution_step_flipped (unit, r, unit->cl->lit->var);
      tmp->idx = r->idx;
      tmp->tr = copy_trace(r->tr);
      delete_rule(r);
      rules[i] = tmp;
      if(is_unit_clause(tmp->cl)) {
        push_unit(tmp);
      }
      else if(tmp->cl && tmp->cl->size == 2) {

        push_pair(tmp);
      }
    }
    else if (lit_in_clause == 1){
      Clause * cl = r->cl;
      deallocate_clause(cl);
      r->cl = 0;
    }
  }
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static Polynomial * unflip_all_var(Rule * r){

  Clause * cl = r->cl;
  while(cl){
    Variable * v = cl->lit->var;
    if(is_flipped_variable(v)){
    //  msg("enter with var %s", v->name);
      Variable * w = flip_variable(v);

      Polynomial * flip_poly = generate_flip_poly(w,v);

      Polynomial * p = unflip_poly_p1_by_poly_p2 (r->pl, flip_poly);

    /*  fprintf(stdout,"flipping lead to \n");
      print_polynomial(p, stdout);
      fprintf(stdout,"\n");*/
      if(p) {
        deallocate_polynomial(r->pl);
        r->pl = p;
      }
      deallocate_polynomial(flip_poly);
    }
    cl = cl->next;
  }

  return r->pl;
}

static int first = 1;
static void add_up_and_print_poly(Rule * r1, Rule * r2){
  Polynomial * p1 = unflip_all_var(r1);
  Polynomial * p2 = unflip_all_var(r2);

  Polynomial * rem = reduce_poly_p1_by_poly_p2 (p1, p2, 1);

  if(first){
    fprintf(mapping_file,"# ");
    fprintf(mapping_file, "%s", rem->rest->lm->term->variable->name);
    fprintf(mapping_file," ");
    fprintf(mapping_file, "%s", rem->lm->term->variable->name);
    fprintf(mapping_file,";\n");
    first = 0;
  }

  print_polynomial(rem, equal_file);
  fprintf(equal_file,";\n");
  deallocate_polynomial(rem);
}



static void resolve_pairs(){
  assert(num_pairs%2 ==0);
  while(num_pairs){
    Rule * r1 = pairs[0];
    int match = 0;
    for(unsigned j = 1; j<num_pairs; j++){
      Rule * r2 = pairs[j];
      if(contain_same_variables(r1->cl, r2->cl)){
        add_up_and_print_poly(r1, r2);
        delete_from_pairs(j);
        delete_from_pairs(0);
        match = 1;
        break;


      }
    }
    if(!match) {
    for(unsigned j= 0; j<num_pairs; j++){
      print_rule(pairs[j], stdout);
    }
      die("no pair found");
    }
  }
}



static void unit_propagation(){
  msg("reached unit prop");
  for(unsigned j = 0; j<num_units; j++){
    Rule * unit = units[j];
    unit_propagation_step(unit);
  }
  resolve_pairs();
}

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static Rule * search_for_resolution(Rule * res1, Rule * r){
  resolv_variable = 0;
  Clause * cl = res1->cl;
  Rule * target = 0;

  while(cl){
    int find = 0;
    Lit * l = cl->lit;
    Trace * comp_trace = r->tr;
    target = 0;

    while(comp_trace){
      Rule * tmp_rule = find_rule(0, num_rules-1, comp_trace->cl);
      if(!tmp_rule) {
        die("did not find rule %i of trace in rule %i", comp_trace->cl, r->idx);
      }
      if(res1->idx == tmp_rule->idx) comp_trace = comp_trace->next;
      else{
        int lit_in_clause = find_lit_in_clause(l, tmp_rule->cl);
        if(lit_in_clause == 1)  find = -1;
        else if(lit_in_clause == -1){
          find++;
          target = tmp_rule;
          if(find > 1) find = -1;
        }
        if(find == -1) {
          comp_trace = 0;
          target = 0;
        }
        else comp_trace = comp_trace ->next;
      }
    }
    if(target != 0 && find == 1){
      resolv_variable = l->var;
      return target;
    }
    cl = cl->next;
  }
  return 0;

}


static Clause * get_next_clause_for_resolution(Rule * r, int size){

  Trace * t = r->tr;
  while(t){
    Rule * current = find_rule(0, num_rules-1, t->cl);
    if(!current) die("nothing found here");
    if(current->cl->size == size){
      Rule * resolvent = search_for_resolution(current, r);
      if(resolvent){
        Rule * tmp = do_resolution_step_flipped (current, resolvent, 0);
        push_rule(tmp);
        r->tr = add_to_trace(r->tr, tmp->idx);

        Trace * del_trace = delete_from_trace(r->tr, resolvent->idx);
        Trace * del2 = delete_from_trace(del_trace, current->idx);

        r->tr = del2;

        if(resolvent->idx > rules[end_traces-1]->idx){delete_from_rules(resolvent);}
        if(current->idx > rules[end_traces-1]->idx){delete_from_rules(current);}

        if(tmp->cl) return copy_clause(tmp->cl);
        else return 0;
      }
    }
    t = t->next;
  }
  return 0;

}



/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static Rule * derive_resolution_proof(Rule * r){
//  msg("resolving on rule %i", r->idx);
  int size = 2;

  while(true){
    Clause * res = get_next_clause_for_resolution(r, size);

    if(!res) size++;
    else if(res->size < size) size = res->size;


    if(!r->tr->next) {
      Clause * cl = r->cl;
      r->cl = res;
      deallocate_clause(cl);
      set_poly_in_rule(r);
      clean_up_rules();
      if(is_bad_target_clause(r)) return r;
      else return 0;
    }

    deallocate_clause(res);
  }
}


static void generate_pac_proof (){
  msg("generate_pac_proof");
  if(start_traces == end_traces) return;
  for(size_t i = start_traces; i < end_traces; i++){
    Rule * r = rules[i];
    Rule * res = derive_resolution_proof(r);
    if(res) {
      msg("resolved opposite of assumption");
      push_unit(r);
      return;
    }
  }
}


/*------------------------------------------------------------------------*/

static void set_all_orig_poly_and_print_to_file(){
  msg("set_all_orig_poly_and_print_to_file");
  for(size_t i = 0; i < num_rules; i++){
    Rule * r = rules[i];
    if(!r->tr) {
      set_poly_in_rule(r);

      bit_flip_rule(r);
    } else {
      if(!start_traces) {
        start_traces = i;
        end_traces = num_rules;
        rule_count = rules[num_rules-1]->idx+1;
      }
      break;
    }
  }
}


/*#############################################################################*/

// PARSING
static size_t size_buffer;
static size_t num_buffer;
static char * buffer;

static void enlarge_buffer () {
  size_t new_size_buffer = size_buffer ? 2*size_buffer : 1;
  REALLOCATE (buffer, size_buffer, new_size_buffer);
  size_buffer = new_size_buffer;
}

static void push_buffer (char ch) {
  if (size_buffer == num_buffer) enlarge_buffer ();
  buffer[num_buffer++] = ch;
}

static void clear_buffer () { num_buffer = 0; }

static void deallocate_buffer () { DEALLOCATE (buffer, size_buffer); }

/*------------------------------------------------------------------------*/

// Character reading routines.

static uint64_t lineno;
static uint64_t columno;
static uint64_t charno;
static const char * parse_file_name;
static FILE * parse_file;
static int close_parse_file;
static int saved_char;
static int char_saved;
static uint64_t last_line_length;

static int next_char () {
  int res;
  if (char_saved) {
    res = saved_char;
    char_saved = 0;
  } else {
#ifdef HAVEUNLOCKEDIO
    res = getc_unlocked (parse_file);
#else
    res = getc (parse_file);
#endif
  }
  if (res == '\n') {
    last_line_length = columno;
    columno = 0;
    lineno++;
  } else if (res != EOF) {
    columno++;
  }
  if (res != EOF) charno++;
  return res;
}

static void prev_char (int ch) {
  assert (!char_saved);
  if (ch == '\n') {
    columno = last_line_length;
    assert (lineno > 0);
    lineno--;
  } else if (ch != EOF) {
    assert (charno > 0);
    charno--;
    assert (columno > 0);
    columno--;
  }
  saved_char = ch;
  char_saved = 1;
}

/*------------------------------------------------------------------------*/
// Scanner.


typedef const char * Token;

static Token END_OF_FILE_TOKEN = "end-of-file";
static Token MINUS_TOKEN       = "minus operator";
static Token SEPARATOR_TOKEN   = "separator";
static Token NUMBER_TOKEN      = "number";
static Token WORD_TOKEN        = "word";

static Token token;
static uint64_t lineno_at_start_of_last_token;
static uint64_t columno_at_start_of_last_token;

static void parse_error (const char * msg, ...) {
  fflush (stdout);
  fprintf (stderr,
    "*** 'drup2dpac' parse error in '%s'"
    " line %" PRIu64,
    parse_file_name,
    lineno_at_start_of_last_token);
  if (buffer[0] && isprint (buffer[0]))
    fprintf (stderr, " at '%s'", buffer);
  else if (token == END_OF_FILE_TOKEN)
    fputs (" at end-of-file", stderr);
  fputs (": ", stderr);
  va_list ap;
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
  exit (1);
}

static Token new_token (Token t) {
  push_buffer (0);
  token = t;
  return token;
}

static Token next_token () {
  clear_buffer ();

  for (;;) {
    int ch = next_char ();
    if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
      continue;
    lineno_at_start_of_last_token = lineno;
    columno_at_start_of_last_token = columno;
    if (ch == EOF) return new_token (END_OF_FILE_TOKEN);
    push_buffer (ch);
    if (ch == '-') return new_token (MINUS_TOKEN);
    if ('0' == ch) {
      return new_token (SEPARATOR_TOKEN);
    }
    if ('1' <= ch && ch <= '9') {
      while ('0' <= (ch = next_char ()) && ch <= '9')
      	push_buffer (ch);
      prev_char (ch);
      return new_token (NUMBER_TOKEN);
    }
    if (is_valid_letter (ch)) {
      while (is_valid_letter (ch = next_char ()))
         push_buffer (ch);
      prev_char (ch);
      return new_token (WORD_TOKEN);
    }

    if (isprint (ch)) parse_error ("invalid character");
    else parse_error ("invalid character code 0x%02x", ch);
  }
}


/*------------------------------------------------------------------------*/

static uint64_t char_to_int () {
  uint64_t res = 0;
  for (const char * p = buffer; *p; p++) {
    if (UINT64_MAX/10 < res)
      parse_error ("invalid number");
    res *= 10;
    int ch = *p;
    assert ('0' <= ch);
    assert (ch <= '9');
    uint64_t digit = ch - '0';
    res += digit;
  }
  return res;
}

static void init_parsing (const char * file_name) {
  assert(file_name);

  parse_file_name = file_name;
  parse_file = fopen (file_name, "r");
  if (!parse_file)
      die ("can not open '%s' for reading", file_name);
  close_parse_file = 1;

  lineno = 1;
  columno = 0;
  charno = 0;
}

static void reset_parsing () {
  if (close_parse_file && fclose (parse_file))
    die ("failed to close '%s'", parse_file_name);
  msg ("read %" PRIu64 " bytes from '%s'", charno, parse_file_name);
}



static Clause * parse_clause(int flag){
  if(flag) next_token();

  while(token != SEPARATOR_TOKEN){
    int sign = 0;
    if(token == MINUS_TOKEN){
      sign = 1;
      next_token();
    }
    assert(token == NUMBER_TOKEN);
    Lit * l = new_lit(sign, buffer, 0);
    push_lit(l);
    next_token();
  }
  Clause * c = build_clause();
  return c;
}

static Trace * parse_trace(){
  next_token();
  while(token != SEPARATOR_TOKEN){
    assert(token == NUMBER_TOKEN);
    uint64_t lit = char_to_int();
    if(lit != bad_clause) {
      push_trace(lit);
    }
    next_token();
  }
  Trace * t = build_trace();
  return t;
}

static void parse_header () {
  next_token();
  assert (token == WORD_TOKEN);
  next_token();
  assert (token == WORD_TOKEN);
  next_token();
  assert (token == NUMBER_TOKEN);
  next_token();
  assert (token == NUMBER_TOKEN);
}


static Rule * parse_init_rule () {
  Clause * c = parse_clause(0);

  assert(token == SEPARATOR_TOKEN);
  Rule * r = new_rule(++init_clauses, c, 0, 0);

  if(is_unit_clause(c)) {
    if(bad_clause) die("cannot have two target clauses");
    msg("set bad clause to %i", init_clauses);
    bad_clause = init_clauses;
    bad_target = r;
  }

  return r;
}


static Rule * parse_rule () {
  assert (token == NUMBER_TOKEN);
  uint64_t idx = char_to_int ();
  Clause * c = parse_clause(1);
  Trace * t = parse_trace();


  assert(token == SEPARATOR_TOKEN);
  Rule * r = new_rule(idx, c, t, 0);

  return r;
}



static void parse_init_clauses (const char * file_name) {
  init_parsing (file_name);
  msg ("reading init_clauses from '%s'", parse_file_name);
  parse_header();

  while(next_token()!=END_OF_FILE_TOKEN){
    Rule * r = parse_init_rule();
    push_rule(r);
  }
  reset_parsing ();

}



static void parse_proof (const char * file_name) {
  init_parsing (file_name);
  msg ("reading proof rules from '%s'", parse_file_name);


  while(next_token()!=END_OF_FILE_TOKEN){
    Rule * r = parse_rule();
    if(r->tr) push_rule(r);
    else(delete_rule(r));
  }
  reset_parsing ();

}
/*------------------------------------------------------------------------*/
// Statistics for max used memory and used process time

// determine max used memory
static size_t maximum_resident_set_size () {
  struct rusage u;
  if (getrusage (RUSAGE_SELF, &u)) return 0;
  return ((size_t) u.ru_maxrss) << 10;
}

// get process time
static double process_time () {
  struct rusage u;
  if (getrusage (RUSAGE_SELF, &u)) return 0;
  double res = u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
  return res;
}

// print statistics of maximum memory and used process time
// print used memory and time
static void print_statistics (double parsing, double init, double proof_gen, double unit_prop) {
  msg ("");
  msg ("maximum resident set size:                %12.2f MB",
    maximum_resident_set_size () / (double)(1<<20));
  msg ("used time for parsing:                    %12.2f seconds", parsing);
  msg ("used time for init:                       %12.2f seconds", init);
  msg ("used time for proof generation:           %12.2f seconds", proof_gen);
  msg ("used time for unit propagation:           %12.2f seconds", unit_prop);
  msg ("used process time:                        %12.2f seconds", process_time ());
}
/*------------------------------------------------------------------------*/
static void deallocate_all () {

  deallocate_lits ();
  deallocate_traces ();
  deallocate_buffer ();
  deallocate_rules();
  deallocate_units();
  deallocate_pairs();
  deallocate_variables();
  deallocate_vstack ();
  deallocate_mstack ();
  deallocate_terms  ();
  assert (!current_allocated_bytes);
}

static void reset () {
  deallocate_all ();
}
/*------------------------------------------------------------------------*/

// Main routine for stand alone checker.

static void banner () {
  msg ("DRUP2DPAC Version " VERSION);
  msg ("Translating DRUP proofs to DPAC");
  msg ("Copyright (C) 2019, Daniela Kaufmann, Johannes Kepler University Linz");
}

int main (int argc, char ** argv) {
  for (int i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-h") ||
        !strcmp (argv[i], "--help"))
      fputs (usage, stdout), exit (0);
    else if (argv[i][0] == '-')
      die ("invalid command line option '%s' (try '-h')", argv[i]);

    else if (output_file_name4)
      die ("too many command line arguments (try '-h')");
    else if (output_file_name3) output_file_name4 = argv[i];
    else if (output_file_name2) output_file_name3 = argv[i];
    else if (output_file_name) output_file_name2 = argv[i];
    else if (input_file_name2) output_file_name = argv[i];
    else if (input_file_name)  input_file_name2 = argv[i];
    else input_file_name = argv[i];
  }

  if(!input_file_name) die("no input file given - check '-h'");
  if(!output_file_name) die("no output file given - check '-h'");
  if(!output_file_name4) die("too few output files given - check '-h'");


  if (!(polys_file = fopen (output_file_name, "a")))
    die ("can not write output to '%s'", output_file_name);

  if (!(pac_file = fopen (output_file_name2, "a")))
    die ("can not write output to '%s'", output_file_name2);

  if (!(equal_file = fopen (output_file_name3, "w")))
      die ("can not write output to '%s'", output_file_name3);

  if (!(mapping_file = fopen (output_file_name4, "a")))
      die ("can not write output to '%s'", output_file_name4);

  banner ();
  init_nonces();
  parse_init_clauses(input_file_name);

  parse_proof(input_file_name2);
  double parse_t = process_time();

  set_all_orig_poly_and_print_to_file();
  double init_t = process_time() - parse_t;

  generate_pac_proof();
  double proof_t = process_time() - parse_t - init_t;

  unit_propagation();
  double unit_t = process_time() - parse_t - init_t - proof_t;

  print_statistics(parse_t, init_t, proof_t, unit_t);
  reset();

  fclose(polys_file);
  fclose(pac_file);
  fclose(equal_file);
  fclose(mapping_file);

  return 0;
}
